#include <visenv.h>

#define REGNUM 50
/***************************************************************
 * Recursive Region Growing Method (from Dr. Bruce A. Maxwell) *
 * Jiwon Shin 09/26/03                                         *
 ***************************************************************/
int **growRegion(Image *image, int *numRegions, 
		   Pixel bgval, int minRegionSize) {
  int curRegionID, regionIndex[REGNUM], next, **regionMap;
  Pixel *regionColor;
  int i, j, cLoc, cRow, cCol, tRow, tCol;
  Stack *S = initStack();
  int regionSize[REGNUM];
  int debugging=0;
  
  printf("imagesize: %ld\n",image->imagesize);
  // allocate memory for regionMap
  regionMap = (int **)malloc(sizeof(int *));
  regionMap[0] = (int *)malloc(sizeof(int)*image->imagesize);

  // initialize regionSize and regionIndex
  for(i=0; i<REGNUM; i++) {
    regionSize[i]=0;
    regionIndex[i]=0;
  }

  // initialize the map to -1
  for(i=0; i<image->imagesize; i++) {
    regionMap[0][i] = -1;
  }

  curRegionID=0;
  for(i=0; i<image->imagesize; i++) {
    // if background, set map = 0
    if(checkPixel(image,i,bgval)) {
      regionMap[0][i] = 0;
    }
    else {    // if foreground
      // if it is not already set, do region growing
      if(regionMap[0][i] == -1) {
	curRegionID++;

	// set map to current region ID
	regionMap[0][i] = curRegionID;
	regionSize[curRegionID]++;
	push(S, i); debugging++;

	while(!isEmpty(S)) {
	  cLoc = pop(S); debugging--;
	  cRow = cLoc/image->cols;
	  cCol = cLoc%image->cols;
	  // test four neighboring pixels
	  for(j=0; j<4; j++) {
	    if((j==0 && cRow<=0) || (j==1 && cCol<=0) ||
	       (j==2 && cCol>=image->cols-1) ||
	       (j==3 && cRow>=image->rows-1)) {
	      break;
	    }
	    tRow = j==0 ? cRow-1 : j==3 ? cRow+1 : cRow;
	    tCol = j==1 ? cCol-1 : j==2 ? cCol+1 : cCol;
	    cLoc = tRow*image->cols + tCol;
	    // if map is not set and is in the same region as the test pixel
	    if(regionMap[0][cLoc]==-1 && 
	       checkPixel(image,cLoc,image->image[i])) {
	      regionMap[0][cLoc] = curRegionID;
	      regionSize[curRegionID]++;
	      push(S, cLoc); debugging++;
	    }
	  }
	}

      }
    }// end else

  }
  
  // set small regions to zero
  next = 1;
  for(i=0; i<REGNUM; i++) {
    if(regionSize[i]) { 
      if(regionSize[i]<minRegionSize) {
	regionSize[i] = 0;
	curRegionID--;
      }
      else {
	regionIndex[i] = next;
	next++;
      }
    }
  }

  // remove small regions
  for(i=0; i<image->imagesize; i++) {
    if(regionMap[0][i]) {
      if(regionSize[regionMap[0][i]]==0) {
	regionMap[0][i]=0;
      }
    }
  }

  // re-label the regionMap
  for(i=0; i<image->imagesize; i++) {
    if(regionMap[0][i]) {
      regionMap[0][i] = regionIndex[regionMap[0][i]];
    }
  }

  // randomly pick region colors
  regionColor = (Pixel *)malloc(sizeof(Pixel)*(curRegionID+1));
  regionColor[0] = (Pixel){0,0,0};
  for(i=1; i<curRegionID+1; i++) {
    regionColor[i].r = drand48()*200+55;
    regionColor[i].g = drand48()*200+55;
    regionColor[i].b = drand48()*200+55;
  }

  // set image to image regions
  for(i=0; i<image->imagesize; i++) {
    image->image[i] = regionColor[regionMap[0][i]];
  }  

  *numRegions = curRegionID;
  free(S);
  free(regionColor);

  return regionMap;
}

/*******************************************************************
 * Find the centroid, major and minor axis, oriented bounding box, *
 * projection to the major and minor axis, and percent filled.     *
 * Jiwon Shin 10/21/03                                             *
 *******************************************************************/
BBox **boundingBox(int *map, int rows, int cols, int numRegions, int *maxRegID) {
  BBox **bbox;
  float m00, m10, m01, bbarea, *area;
  double alpha, beta, u11, u02, u20;
  double temp1, temp2;
  int i, j, k, maxReg, maxID=0;
  long loc;

  // allocating memory
  bbox = (BBox **)malloc(sizeof(BBox *));
  bbox[0] = (BBox *)malloc(sizeof(BBox)*numRegions);
  area = (float *)malloc(sizeof(float)*numRegions);

  // initialize bbox
  for(i=0; i<numRegions; i++) {
    bbox[0][i].centroid[0] = bbox[0][i].centroid[1] = 0;
    bbox[0][i].bbox[0] = bbox[0][i].bbox[1] = 1000;
    bbox[0][i].bbox[2] = bbox[0][i].bbox[3] = 0;
  }

  // find the bounding boxes
  for(i=0; i<rows; i++) {
    for(j=0; j<cols; j++) {
      loc = i*cols+j;
      // if not background
      if(map[loc]) {
	bbox[0][map[loc]-1].bbox[0] = (i < bbox[0][map[loc]-1].bbox[0] ? 
				       i : bbox[0][map[loc]-1].bbox[0]); 
	bbox[0][map[loc]-1].bbox[1] = (j < bbox[0][map[loc]-1].bbox[1] ? 
				       j : bbox[0][map[loc]-1].bbox[1]); 
	bbox[0][map[loc]-1].bbox[2] = (i > bbox[0][map[loc]-1].bbox[2] ? 
				       i : bbox[0][map[loc]-1].bbox[2]); 
	bbox[0][map[loc]-1].bbox[3] = (j > bbox[0][map[loc]-1].bbox[3] ? 
				       j : bbox[0][map[loc]-1].bbox[3]); 
      }
    }
  }

  maxReg=0;
  // calculate centroid
  for(i=0; i<numRegions; i++) {
    m00 = m10 = m01 = 0.0;
    for(j=bbox[0][i].bbox[0]; j<=bbox[0][i].bbox[2]; j++) {
      for(k=bbox[0][i].bbox[1]; k<=bbox[0][i].bbox[3]; k++) {
	if(map[j*cols+k]==(i+1)) {
	  m00++;
	  m10 += k;
	  m01 += j;
	}
      }
    }
    printf("m00 %f\tm10 %f\tm01 %f\n",m00,m10,m01);   
    bbox[0][i].centroid[0] = (long)(m10/m00+.5); //x
    bbox[0][i].centroid[1] = (long)(m01/m00+.5); //y
    area[i]=m00;
    if(m00>maxReg) {
      maxReg=m00;
      maxID=i;
    }
    //printf("%d:: centroid %ld\t%ld\n",i,bbox[0][i].centroid[0],bbox[0][i].centroid[1]);
  }
  *maxRegID=maxID;

  printf("central axes\n");
  // use 2nd central moment to determine the angle of the axes
  for(i=0; i<numRegions; i++) {
    u11 = u02 = u20 = 0.0;
    for(j=bbox[0][i].bbox[0]; j<=bbox[0][i].bbox[2]; j++) {
      for(k=bbox[0][i].bbox[1]; k<=bbox[0][i].bbox[3]; k++) {
	if(map[j*cols+k]==(i+1)) {
	  temp1 = j-bbox[0][i].centroid[1]; //row
	  temp2 = k-bbox[0][i].centroid[0]; //col
	  u11 += temp1*temp2;
	  u20 += temp2*temp2; //x
	  u02 += temp1*temp1; //y
	}
      }
    }
    printf("%d:: numRegions %d\n",i,numRegions);
    // finding the major and the minor axis
    printf("u11 %f\tu02 %f\tu20 %f\n",u11,u02,u20);
    alpha = atan2(2.0*u11, u02-u20)/2;    beta = alpha + M_PI/2;
    printf("alpha %f\t beta %f\n",alpha,beta);
    centralProjection(alpha, beta, map, rows, cols, i+1, bbox[0][i].centroid, 
		      bbox[0][i].bbox, bbox[0][i].axis, bbox[0][i].obox, 
		      bbox[0][i].proj);

    bbarea = sqrt((pow((bbox[0][i].obox[0][0]-bbox[0][i].obox[1][0]),2)+
		   pow((bbox[0][i].obox[0][1]-bbox[0][i].obox[1][1]),2))) *
	     sqrt((pow((bbox[0][i].obox[1][0]-bbox[0][i].obox[2][0]),2)+
		   pow((bbox[0][i].obox[1][1]-bbox[0][i].obox[2][1]),2)));
    bbox[0][i].percentFilled=area[i]/bbarea;
    printf("percentFilled %f\n",bbox[0][i].percentFilled);
  }

  free(area);
  return bbox;
}

void centralProjection(double alpha, double beta, int *map, int rows, int cols,
		       int id, Point_t centroid, Box_t bbox, Point axis[4], 
		       Point obox[4], Projection_t proj[2]) {
  
  float at1, at2, bt1, bt2; //temp
  float ad1, ad2, bd1, bd2; //delta
  float delta, d0[2], d1[2];
  long cX, cY;
  int i, j, cons[2];
  int *temp;

  at1 = cos(alpha);  at2 = -sin(alpha);
  bt1 = cos(beta);   bt2 = -sin(beta);
  
  // determining delta
  if( (at1>0 && at2>0) || (at1<0 && at2<0) ) { 
    at1 = fabs(at1); at2=fabs(at2);
    ad1 = (abs((centroid[0]-bbox[1])/at1) < abs((centroid[1]-bbox[0])/at2) ? 
	   abs((centroid[0]-bbox[1])/at1) : abs((centroid[1]-bbox[0])/at2));
    
    ad2 = (abs((bbox[3]-centroid[0])/at1) < abs((bbox[2]-centroid[1])/at2) ? 
	   abs((bbox[3]-centroid[0])/at1) : abs((bbox[2]-centroid[1])/at2));
  }
  else {
    at1 = -at1; at2 = -at2;
    ad1 = (abs((centroid[0]-bbox[3])/at1) < abs((centroid[1]-bbox[0])/at2) ? 
	   abs((centroid[0]-bbox[3])/at1) : abs((centroid[1]-bbox[0])/at2));
    
    ad2 = (abs((bbox[1]-centroid[0])/at1) < abs((bbox[2]-centroid[1])/at2) ? 
	   abs((bbox[1]-centroid[0])/at1) : abs((bbox[2]-centroid[1])/at2));
  }

  if( (bt1>0 && bt2>0) || (bt1<0 && bt2<0) ) { 
    bt1 = fabs(bt1); bt2=fabs(bt2);
    bd1 = (abs((centroid[0]-bbox[1])/bt1) < abs((centroid[1]-bbox[0])/bt2) ? 
	   abs((centroid[0]-bbox[1])/bt1) : abs((centroid[1]-bbox[0])/bt2));
    
    bd2 = (abs((bbox[3]-centroid[0])/bt1) < abs((bbox[2]-centroid[1])/bt2) ? 
	   abs((bbox[3]-centroid[0])/bt1) : abs((bbox[2]-centroid[1])/bt2));
  }
  else {
    bt1 = -bt1; bt2 = -bt2;
    bd1 = (abs((centroid[0]-bbox[3])/bt1) < abs((centroid[1]-bbox[0])/bt2) ? 
	   abs((centroid[0]-bbox[3])/bt1) : abs((centroid[1]-bbox[0])/bt2));
    
    bd2 = (abs((bbox[1]-centroid[0])/bt1) < abs((bbox[2]-centroid[1])/bt2) ? 
	   abs((bbox[1]-centroid[0])/bt1) : abs((bbox[2]-centroid[1])/bt2));
  }

  // projecting to central axis
  proj[0].size = ad1 > ad2+1 ? 2*(ad1+10)+1 : 2*(ad2+11)+1;
  proj[0].p = (int *)calloc(proj[0].size,sizeof(int));
  printf("proj[0].size = %d\tat1 %f\tat2 %f\n",proj[0].size,at1,at2);

  cons[0] = (proj[0].size-1)/2; delta = bd1 > bd2+1 ? bd1+10 : bd2+11;
  for(i=0; i<proj[0].size; i++) {
    cX = centroid[0]+(cons[0]-i)*at1;
    cY = centroid[1]+(cons[0]-i)*at2;
    proj[0].p[i]=countIntersections(cX,cY,delta,bt1,bt2,map,id,rows,cols);
  }

  proj[1].size = bd1 > bd2+1 ? 2*(bd1+10)+1 : 2*(bd2+11)+1;
  proj[1].p = (int *)calloc(proj[1].size,sizeof(int));
  printf("proj[1].size = %d\tbt1 %f\tbt2 %f\n",proj[1].size,bt1,bt2);

  cons[1] = (proj[1].size-1)/2; delta = ad1 > ad2+1 ? ad1+10 : ad2+11;
  for(i=0; i<proj[1].size; i++) {
    cX = centroid[0]+(cons[1]-i)*bt1;
    cY = centroid[1]+(cons[1]-i)*bt2;
    proj[1].p[i]=countIntersections(cX,cY,delta,at1,at2,map,id,rows,cols);
  }

  for(j=0; j<2; j++) {
    printf("cons[%d]: %d\n",j,cons[j]);
    for(i=0; i<proj[j].size; i++) {
      if(proj[j].p[i]>0) {
	d0[j]=cons[j]-i;
	printf("%d\td0[%d]: %f\n",i,j,d0[j]);
	break;
      }
    }
    for(i=proj[j].size-1; i>=0; i--) {
      if(proj[j].p[i]>0) {
	d1[j]=i-cons[j];
	printf("%d\td1[%d]: %f\n",i,j,d1[j]);
	break;
      }
    }
  }

  // resize projections
  temp = resizeHistogram(proj[0].p,proj[0].size);
  free(proj[0].p);
  proj[0].p = temp; proj[0].size=100;
  
  temp = resizeHistogram(proj[1].p,proj[1].size);
  free(proj[1].p);
  proj[1].p = temp; proj[1].size=100;

  // calculate the location of the four edge points of the oriented bounding box
  // and the central axes
  cX = centroid[0]+d0[0]*at1;
  cY = centroid[1]+d0[0]*at2;
  
  printf("cX: %ld\tcY: %ld\n",cX,cY);
  obox[0][0] = cX+d0[1]*bt1;
  obox[0][1] = cY+d0[1]*bt2;
  obox[3][0] = cX-d1[1]*bt1;
  obox[3][1] = cY-d1[1]*bt2;

  printf("pt0[0]: %f\tpt0[1]: %f\n",obox[0][0],obox[0][1]);
  printf("pt3[0]: %f\tpt3[1]: %f\n",obox[3][0],obox[3][1]);
  axis[0][0] = cX;
  axis[0][1] = cY;

  cX = centroid[0]-d1[0]*at1;
  cY = centroid[1]-d1[0]*at2;
  
  printf("cX: %ld\tcY: %ld\n",cX,cY);
  obox[1][0] = cX+d0[1]*bt1;
  obox[1][1] = cY+d0[1]*bt2;
  obox[2][0] = cX-d1[1]*bt1;
  obox[2][1] = cY-d1[1]*bt2;
  printf("pt1[0]: %f\tpt1[1]: %f\n",obox[1][0],obox[1][1]);
  printf("pt2[0]: %f\tpt2[1]: %f\n",obox[2][0],obox[2][1]);

  axis[1][0] = cX;
  axis[1][1] = cY;

  axis[2][0] = centroid[0]+d0[1]*bt1;
  axis[2][1] = centroid[1]+d0[1]*bt2;

  axis[3][0] = centroid[0]-d1[1]*bt1;
  axis[3][1] = centroid[1]-d1[1]*bt2;
}

int countIntersections(long sx, long sy, float d, float dx, float dy, int *map, 
		       int id, int rows, int cols) {
  
  Point start, end;
  long xLoc, yLoc, Loc, count;
  long i, ldx, ldy;
  float x, y;
  long temp, interchange, e;
  long s1, s2;
  
  start[0]=sx+d*dx;
  start[1]=sy+d*dy;
  end[0]=sx-d*dx;
  end[1]=sy-d*dy;
    
  xLoc = start[0];
  yLoc = start[1];
  
  ldx = abs(end[0] - start[0]);
  ldy = abs(end[1] - start[1]); 
  s1 = sign(end[0], start[0]);
  s2 = sign(end[1], start[1]);
  x = start[0] - .5 * s1;
  y = start[1] - .5 * s2;
  
  if( ldy > ldx ) {
    temp = ldx;
    ldx = ldy;
    ldy = temp;
    interchange = 1;
  } else {
    interchange = 0;
  }  
  e = 3 * ldy - 2 * ldx;
  
  count=0;
  for(i=0; i<ldx; i++) {
    Loc = yLoc*cols+xLoc;
    if(yLoc>=0 && xLoc>=0 && yLoc<rows && xLoc<cols) {
       if(map[Loc]==id) {
	 count++;
       }
    }
    while( e > 0 ) {
      if( interchange == 1 ) {
	x = x + s1;
	xLoc = x + .5 * s1;
      }
      else {
	y = y + s2;
	yLoc = y + .5 * s2;
      }
      e = e - 2 * ldx;
    }
    if( interchange == 1 ) {
      y = y + s2;
      yLoc = y + .5 * s2;
    }
    else {
      x = x + s1;
      xLoc = x + .5 * s1;
    }
    e = e + 2 * ldy;
  }
    
  return count;
}
